#!/usr/bin/env python
'''
  #############################################################################
  #
  #     This Cplant(TM) source code is the property of Sandia National
  #     Laboratories.
  #
  #     This Cplant(TM) source code is copyrighted by Sandia National
  #     Laboratories.
  #
  #     The redistribution of this Cplant(TM) source code is subject to the
  #     terms of the GNU Lesser General Public License
  #     (see cit/LGPL or http://www.gnu.org/licenses/lgpl.html)
  #
  #     Cplant(TM) Copyright 1998, 1999, 2000, 2001, 2002 Sandia Corporation.
  #     Under the terms of Contract DE-AC04-94AL85000, there is a non-exclusive
  #     license for use of this work by or on behalf of the US Government.
  #     Export of this program may require a license from the United States
  #     Government.
  #
  #############################################################################

 File    : apitest.py
 Revision: $Revision: 1.7 $
 Author  : William McLendon, Narayan Desai
 Description:
	APITest implementation using miniDOM processing rather than a SAX parser.
	Currently under test.
'''
import sys
from os import path
from xml.dom.minidom import parse, Document
from getopt import getopt,GetoptError
from framework import TestSet
from errorhandlers import TestSetupError
from md5 import md5

def getFileMD5(filename):
    IFP = open(filename,'r')
    text = ""
    for line in IFP:
        text += line
    md5sum = md5(text)
    return md5sum.hexdigest() 

def usage():
    print "Usage:\n\tapitest.py -f <filename> [-d --test --sqldb]"

# --------------------
#  main
# --------------------
if __name__ == '__main__':
    file = ""

    # get the command line options
    try:
        (opts,args)=getopt(sys.argv[1:],'dvf:l:q', \
                    ['file=','testmsg','sqldb','verbose','logfile=','sqluser=','sqlpw=','quiet', \
                     'test=','noAncestors','noDescendents','resultfile='])
    except GetoptError,msg:
        print msg
        usage()
        sys.exit(1)

    options = {}
    # Set defaults
    options['sqluser'] = 'apiuser'
    options['sqlpw'] = 'apipass'

    # set flags based on command line options
    for (opt,data) in opts:
        if opt == '-f' or opt=='--file':
            options['filename'] = data
        elif opt == '-d':
            options['debug'] = 1
        elif opt == '--testmsg':
            options['testmsg'] = 1
        elif opt == '--sqldb':
            options['sql'] = 1
        elif opt == '-v' or opt=='--verbose':
            options['verbose'] = 1
        elif opt == '-l' or opt=='--logfile':
            options['logfile'] = data
        elif opt == '--sqluser':
            options['sqluser'] = data
        elif opt == '--sqlpw':
            options['sqlpw'] = data
        elif opt == '--quiet' or opt=='-q':
            options['quiet'] = 1
        elif opt == '--resultfile':
            options['resultfile'] = data
        elif opt[:2]=='--':
            options[opt[2:]] = data

    # Debugging: display received command line options
    if options.has_key('debug'):
        print "D:\tCommand Line Options:", options

    # Print out a message to stdout and stderr for testing (used to test APITEST)
    if options.has_key('testmsg'):
        print >>sys.stdout,"This is a stdout message"
        print >>sys.stderr,"This is a stderr message"
        sys.stdout.flush()
        sys.stderr.flush()
        sys.exit(20)

    # if input file name is not provided, print usage and exit with an error.
    if not options.has_key('filename'):
        usage()
        sys.exit(1)

    # Verfiy existence of input file
    if not path.exists(options['filename']):
        print "I'm sorry, I could not locate '%s'."%(options['filename'])
        sys.exit(1)
    elif not path.isfile(options['filename']):
        print "I'm sorry, '%s' is not a file."%(options['filename'])
        sys.exit(1)

    # Parse the XML file 
    try:
        data=parse(options['filename'])
    except:
        print "I'm sorry, parsing '%s' failed."%(options['filename'])
        sys.exit(2)

    # All is well, execute the tests
    test=TestSet(data.documentElement)
    test.execute(options)

    if options.has_key('debug'): print "MD5('%s')="%options['filename'], getFileMD5(options['filename'])

# EOF
