#! /usr/bin/env python

import atexit
from getopt import getopt, GetoptError
from xml.dom.minidom import Document, parseString
from os import environ, getuid, fork
from time import sleep
from pwd import getpwuid
from sss.ssslib import comm_lib, LookupError
from sss.server import SSSSelectEventReceiver
from string import join, split, strip
from sys import argv, exit

class procevent(SSSSelectEventReceiver):
    def __init__(self, pgid, debug=0):
        self.dispatch={'event':self.HandleXMLEvent}
        SSSSelectEventReceiver.__init__(self,component='process-manager',msg='**',data=pgid,debug=debug)

    def HandleXMLEvent(self, xe, (peer,port)):
        for event in xe.childNodes:
            if event.getAttribute("msg") == "process_end":
                self.shut = 1
        response=Document().createElement('event-ok')
        return response
    
if __name__ == '__main__':
    try:
        (opts,args)=getopt(argv[1:], 'aw:g:f:d')
    except GetoptError, msg:
        print "Usage: dsh.py [-a] [-d] [-g group] [-f file] [-w node1,node2] cmd"
        print msg
        exit(1)

    mode=None
    for opt in opts:
        if opt[0] == '-a':
            mode='all'
        elif opt[0] == '-g':
            mode='group'
            tag=opt[1]
        elif opt[0] == '-f':
            mode='node'
            nodes=map(strip,open(opt[1]).readlines())
        elif opt[0] == '-w':
            mode='node'
            nodes=split(opt[1],',')
        elif opt[0] == '-d':
            pass
        else:
            print "semantic error"
            exit(1)

    if not args:
        print "Usage: dsh.py [-a] [-d] [-g group] [-f file] [-w node1,node2] cmd"
        print "One of -a, -g, -f, or -w are required"
        exit(1)

    cmd=args[0]

    args=args[1:]

    if not mode:
        print "Usage: dsh.py [-a] [-d] [-g group] [-f file] [-w node1,node2] cmd"
        print "One of -a, -g, -f, or -w are required"
        exit(1)

    submitter = getpwuid(getuid())[0]
    output = 'display'
    path = environ['PATH']
    cwd = environ['PWD']

    comm = comm_lib(debug=('-d' in argv))

    #if mode=group, query the build-system to find out the list of nodes
    if mode == 'group':
        query = Document()
        wrapper = query.createElement('get-tag')
        newel = query.createElement('tag')
        newel.setAttribute('tag',tag)
        newel.setAttribute('node','*')
        wrapper.appendChild(newel)
        query.appendChild(wrapper)
        bsh = comm.ClientInit('build-system')
        comm.SendMessage(bsh,query.toxml())
        answer = comm.RecvMessage(bsh)
        comm.ClientClose(bsh)
        domobj = parseString(answer)
        # not finished here but I am waiting to test it for real on my machine, wanted to checkin 
        # put the hostlist into nodes
        nodes=map(lambda x:domobj.getElementsByTagName("tag").__getitem__(x).getAttribute('node'),range(0,len(domobj.getElementsByTagName("tag"))) )
    
    # at this point, regardless of mode, nodes is populated
    numprocs=str(len(nodes))
    
    msg = Document().createElement('create-process-group')
    msg.setAttribute('totalprocs',numprocs)
    msg.setAttribute('submitter',submitter)
    msg.setAttribute('output',output)
    ps =  Document().createElement('process-spec')
    ps.setAttribute('exec',cmd)
    ps.setAttribute('path',path)
    ps.setAttribute('cwd',cwd)
    msg.appendChild(ps)
    hs=Document().createElement('host-spec')
    hsd=Document().createTextNode(join(nodes,'\n'))
    hs.appendChild(hsd)
    msg.appendChild(hs)
    for arg in range(len(args)):
        ax=Document().createElement('arg')
        ax.setAttribute('idx',str(arg))
        ax.setAttribute('value',args[arg])
        ps.appendChild(ax)
        
    try:
        pm = comm.ClientInit('process-manager')
    except LookupError:
        print "Process Manager down"
        exit(1)
        
    comm.SendMessage(pm,msg.toxml())
    ack = comm.RecvMessage(pm)
    ack_dom = parseString(ack)
    if ack_dom.documentElement.nodeName == "error":
        print ack
        exit(1)
    pgid=ack_dom.documentElement.getAttribute('pgid')
    comm.ClientClose(pm)
    debug=argv.count('-d')
    server=procevent(pgid,debug=debug)
    server.ServeForever()
    server.Shutdown()
    atexit._exithandlers=[]
    del server
    # Now the process is done; we need to collect output
    pm=comm.ClientInit('process-manager')
    comm.SendMessage(pm, "<get-process-group-info><process-group pgid='%s'/></get-process-group-info>"%pgid)
    resp=comm.RecvMessage(pm)
    r=parseString(resp)
    comm.ClientClose(pm)

    output=''
    err=''
    try:
        output=join(map(lambda x:x.nodeValue,r.documentElement.childNodes[0].getElementsByTagName("output")[0].childNodes),"")
    except:
        pass

    try:
        err=r.documentElement.childNodes[0].getElementsByTagName("error")[0].childNodes[0].nodeValue
    except:
        pass

    if output: print output,
    if err:
        print err,
        exit(1)
