#include <commlib.h>
#include <ssslib.h>
#include <string.h>
#include <stdlib.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <errno.h>

JNIEXPORT void JNICALL Java_commlib_init_1sss(JNIEnv *env,jobject obj,jstring conffile,jint debug){
  char *c_conffile;
  int c_debug;
  jfieldID sss_fid;
  jclass cls;
  SSS sss;
  jclass ExceptionClass;
  char errbuf[MAXERRLENGTH];

  cls = (*env)->GetObjectClass(env,obj);

  sss_fid = (*env)->GetFieldID(env,cls,"sss_ptr","I");
			       
  c_conffile = (char *)((*env)->GetStringUTFChars(env,conffile,0));
  c_debug = debug;

  sss = SSS_Init(c_conffile,c_debug);

  (*env)->ReleaseStringUTFChars(env,conffile,c_conffile);
  
  if (!sss){
    ExceptionClass = (*env)->FindClass(env,"java/lang/Exception");
    (*env)->ThrowNew(env,ExceptionClass,strerror_r(errno,errbuf,MAXERRLENGTH));
  }

  (*env)->SetIntField(env, obj, sss_fid, (jint)sss);
  
}

JNIEXPORT jint JNICALL Java_commlib_SDRegister(JNIEnv *env,jobject obj,jstring service,jstring host,jint port,jstring protocol,jstring schema_version,jstring tier){
  char *c_service;
  char *c_host;
  char *c_protocol;
  char *c_schema_version;
  char *c_tier;
  int rc;
  jfieldID sss_fid;
  jclass cls;
  SSS sss;
  jclass ExceptionClass;
  char errbuf[MAXERRLENGTH];
  int errtype;

  cls = (*env)->GetObjectClass(env,obj);

  sss_fid = (*env)->GetFieldID(env,cls,"sss_ptr","I");
			       
  sss = (SSS)((*env)->GetIntField(env,obj,sss_fid));

  c_service = (char *)((*env)->GetStringUTFChars(env,service,0));
  c_host = (char *)((*env)->GetStringUTFChars(env,host,0));
  c_protocol = (char *)((*env)->GetStringUTFChars(env,protocol,0));
  c_schema_version = (char *)((*env)->GetStringUTFChars(env,schema_version,0));
  c_tier = (char *)((*env)->GetStringUTFChars(env,tier,0));

  rc = SSS_SDRegister(sss,c_service,c_host,port,c_protocol,c_schema_version,c_tier);

  (*env)->ReleaseStringUTFChars(env,service,c_service);
  (*env)->ReleaseStringUTFChars(env,host,c_host);
  (*env)->ReleaseStringUTFChars(env,protocol,c_protocol);
  (*env)->ReleaseStringUTFChars(env,schema_version,c_schema_version);
  (*env)->ReleaseStringUTFChars(env,tier,c_tier);

  if (rc == -1){
    if (sss->sss_errno == SSS_HOST_NOT_FOUND){
      ExceptionClass = (*env)->FindClass(env,"java/net/UnknownHostException");
      strncpy(errbuf,"Host not found",MAXERRLENGTH);
      (*env)->ThrowNew(env,ExceptionClass,errbuf);
    }
    SSS_strerror(sss->sss_errno,errbuf,&errtype);
    switch (errtype){
    case SSS_PARSE_ERROR:
      ExceptionClass = (*env)->FindClass(env,"SSS_ParseException");
      break;
    case SSS_PROTOCOL_ERROR:
      ExceptionClass = (*env)->FindClass(env,"SSS_ProtocolException");
      break;
    case SSS_SOCKET_ERROR:
    case SSS_CONNECT_ERROR:
    case SSS_COMMUNICATION_ERROR:
      ExceptionClass = (*env)->FindClass(env,"java/io/IOException");
      break;
    case SSS_INTERNAL_ERROR:
      ExceptionClass = (*env)->FindClass(env,"SSS_InternalError");
      break;
    }
    (*env)->ThrowNew(env,ExceptionClass,errbuf);
  }

  return rc;
  
}

JNIEXPORT jint JNICALL Java_commlib_SDRemove(JNIEnv *env,jobject obj,jstring service,jstring host,jint port,jstring protocol){
  char *c_service;
  char *c_host;
  char *c_protocol;
  int rc;
  jfieldID sss_fid;
  jclass cls;
  SSS sss;
  jclass ExceptionClass;
  char errbuf[MAXERRLENGTH];
  int errtype;

  cls = (*env)->GetObjectClass(env,obj);

  sss_fid = (*env)->GetFieldID(env,cls,"sss_ptr","I");
			       
  sss = (SSS)((*env)->GetIntField(env,obj,sss_fid));

  c_service = (char *)((*env)->GetStringUTFChars(env,service,0));
  c_service = (char *)((*env)->GetStringUTFChars(env,host,0));
  c_protocol = (char *)((*env)->GetStringUTFChars(env,protocol,0));

  rc = SSS_SDRemove(sss,c_service,c_host,port,c_protocol);

  (*env)->ReleaseStringUTFChars(env,service,c_service);
  (*env)->ReleaseStringUTFChars(env,protocol,c_protocol);

  if (rc == -1){
    if (sss->sss_errno == SSS_HOST_NOT_FOUND){
      ExceptionClass = (*env)->FindClass(env,"java/net/UnknownHostException");
      strncpy(errbuf,"Host not found",MAXERRLENGTH);
      (*env)->ThrowNew(env,ExceptionClass,errbuf);
    }
    SSS_strerror(sss->sss_errno,errbuf,&errtype);
    switch (errtype){
    case SSS_PARSE_ERROR:
      ExceptionClass = (*env)->FindClass(env,"SSS_ParseException");
      break;
    case SSS_PROTOCOL_ERROR:
      ExceptionClass = (*env)->FindClass(env,"SSS_ProtocolException");
      break;
    case SSS_SOCKET_ERROR:
    case SSS_CONNECT_ERROR:
    case SSS_COMMUNICATION_ERROR:
      ExceptionClass = (*env)->FindClass(env,"java/io/IOException");
      break;
    case SSS_INTERNAL_ERROR:
      ExceptionClass = (*env)->FindClass(env,"SSS_InternalError");
      break;
    }
    (*env)->ThrowNew(env,ExceptionClass,errbuf);
  }

  return rc;
}

JNIEXPORT jstring JNICALL Java_commlib_ServerInit(JNIEnv *env,jobject obj,jint port){
  jfieldID sss_fid;
  jclass cls;
  SSS sss;
  jclass ExceptionClass;
  char errbuf[MAXERRLENGTH];
  int errtype;

  int servfd;
  int accepted_fd;
  struct sockaddr_in addr;
  int yes = 1;
  struct sockaddr_in cli_addr;
  socklen_t sin_size;
  char *handle;
  int rc;

  cls = (*env)->GetObjectClass(env,obj);

  addr.sin_family = AF_INET;
  addr.sin_port = htons(port);
  addr.sin_addr.s_addr = INADDR_ANY;
  memset(&(addr.sin_zero), '\0', 8);  

  sss_fid = (*env)->GetFieldID(env,cls,"sss_ptr","I");
			       
  sss = (SSS)((*env)->GetIntField(env,obj,sss_fid));

  servfd = socket(AF_INET,SOCK_STREAM,0);

  if (servfd == -1){
    ExceptionClass = (*env)->FindClass(env,"java/io/IOException");
    (*env)->ThrowNew(env,ExceptionClass,strerror_r(errno,errbuf,MAXERRLENGTH));
  }

  setsockopt(servfd, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(int));

  rc = bind(servfd,(struct sockaddr *)&addr,sizeof(struct sockaddr));

  if (rc == -1){
    ExceptionClass = (*env)->FindClass(env,"java/io/IOException");
    (*env)->ThrowNew(env,ExceptionClass,strerror_r(errno,errbuf,MAXERRLENGTH));
  }

  listen(servfd,5);

  accepted_fd = accept(servfd,(struct sockaddr *)&cli_addr,&sin_size);

  if (accepted_fd == -1){
    ExceptionClass = (*env)->FindClass(env,"java/io/IOException");
    (*env)->ThrowNew(env,ExceptionClass,strerror_r(errno,errbuf,MAXERRLENGTH));
  }

  handle = SSS_ServerInit(sss,accepted_fd);

  if (!handle){
    SSS_strerror(sss->sss_errno,errbuf,&errtype);
    switch (errtype){
    case SSS_PROTOCOL_ERROR:
      ExceptionClass = (*env)->FindClass(env,"SSS_ProtocolException");
      break;
    case SSS_INTERNAL_ERROR:
    case SSS_SD_LOOKUP_ERROR:
      ExceptionClass = (*env)->FindClass(env,"SSS_InternalError");
      break;
    case SSS_SOCKET_ERROR:
    case SSS_CONNECT_ERROR:
    case SSS_COMMUNICATION_ERROR:
      ExceptionClass = (*env)->FindClass(env,"java/io/IOException");
      break;
    }
    (*env)->ThrowNew(env,ExceptionClass,errbuf);
  }

  return (*env)->NewStringUTF(env,handle);
       
}

JNIEXPORT jstring JNICALL Java_commlib_ClientInit(JNIEnv *env,jobject obj,jstring service){
  char *c_service;
  char *ret;
  jstring retval;
  jfieldID sss_fid;
  jclass cls;
  SSS sss;
  jclass ExceptionClass;
  char errbuf[MAXERRLENGTH];
  int errtype;

  cls = (*env)->GetObjectClass(env,obj);

  sss_fid = (*env)->GetFieldID(env,cls,"sss_ptr","I");
			       
  sss = (SSS)((*env)->GetIntField(env,obj,sss_fid));

  c_service = (char *)((*env)->GetStringUTFChars(env,service,0));

  ret = SSS_ClientInit(sss,c_service);

  if (!ret){
    if (sss->sss_errno == SSS_HOST_NOT_FOUND){
      ExceptionClass = (*env)->FindClass(env,"java/net/UnknownHostException");
      strncpy(errbuf,"Host not found",MAXERRLENGTH);
      (*env)->ThrowNew(env,ExceptionClass,errbuf);
    }
    SSS_strerror(sss->sss_errno,errbuf,&errtype);
    switch (errtype){
    case SSS_PARSE_ERROR:
      ExceptionClass = (*env)->FindClass(env,"SSS_ParseException");
      break;
    case SSS_PROTOCOL_ERROR:
      ExceptionClass = (*env)->FindClass(env,"SSS_ProtocolException");
      break;
    case SSS_SOCKET_ERROR:
    case SSS_CONNECT_ERROR:
    case SSS_COMMUNICATION_ERROR:
      ExceptionClass = (*env)->FindClass(env,"java/io/IOException");
      break;
    case SSS_INTERNAL_ERROR:
      ExceptionClass = (*env)->FindClass(env,"SSS_InternalError");
      break;
    }
    (*env)->ThrowNew(env,ExceptionClass,errbuf);
  }

  retval = (*env)->NewStringUTF(env,ret);

  (*env)->ReleaseStringUTFChars(env,service,c_service);

  return retval;
}

JNIEXPORT jint JNICALL Java_commlib_ClientClose(JNIEnv *env,jobject obj,jstring handle){
  char *c_handle;
  int rc;
  jfieldID sss_fid;
  jclass cls;
  SSS sss;
  jclass ExceptionClass;
  char errbuf[MAXERRLENGTH];
  int errtype;
  
  cls = (*env)->GetObjectClass(env,obj);

  sss_fid = (*env)->GetFieldID(env,cls,"sss_ptr","I");
			       
  sss = (SSS)((*env)->GetIntField(env,obj,sss_fid));

  c_handle = (char *)((*env)->GetStringUTFChars(env,handle,0));

  rc = SSS_ClientClose(sss,c_handle);

  (*env)->ReleaseStringUTFChars(env,handle,c_handle);

  if (rc == -1){
    SSS_strerror(sss->sss_errno,errbuf,&errtype);
    switch (errtype){
    case SSS_SOCKET_ERROR:
      ExceptionClass = (*env)->FindClass(env,"java/io/IOException");
      break;
    case SSS_INTERNAL_ERROR:
      ExceptionClass = (*env)->FindClass(env,"SSS_InternalError");
    }
    (*env)->ThrowNew(env,ExceptionClass,errbuf);
  }

  return rc;
}

JNIEXPORT jint JNICALL Java_commlib_ServerClose(JNIEnv *env,jobject obj,jstring handle){
  char *c_handle;
  int rc;
  jfieldID sss_fid;
  jclass cls;
  SSS sss;
  jclass ExceptionClass;
  char errbuf[MAXERRLENGTH];
  int errtype;
  
  cls = (*env)->GetObjectClass(env,obj);

  sss_fid = (*env)->GetFieldID(env,cls,"sss_ptr","I");
			       
  sss = (SSS)((*env)->GetIntField(env,obj,sss_fid));

  c_handle = (char *)((*env)->GetStringUTFChars(env,handle,0));

  rc = SSS_ServerClose(sss,c_handle);

  (*env)->ReleaseStringUTFChars(env,handle,c_handle);

  if (rc == -1){
    SSS_strerror(sss->sss_errno,errbuf,&errtype);
    switch (errtype){
    case SSS_SOCKET_ERROR:
      ExceptionClass = (*env)->FindClass(env,"java/io/IOException");
      break;
    case SSS_INTERNAL_ERROR:
      ExceptionClass = (*env)->FindClass(env,"SSS_InternalError");
    }
    (*env)->ThrowNew(env,ExceptionClass,errbuf);
  }

  return rc;
}

JNIEXPORT jint JNICALL Java_commlib_SendMessage(JNIEnv *env,jobject obj,jstring handle,jstring message){
  char *c_message;
  char *c_handle;
  int rc;
  jfieldID sss_fid;
  jclass cls;
  SSS sss;
  jclass ExceptionClass;
  char errbuf[MAXERRLENGTH];
  int errtype;

  cls = (*env)->GetObjectClass(env,obj);

  sss_fid = (*env)->GetFieldID(env,cls,"sss_ptr","I");
			       
  sss = (SSS)((*env)->GetIntField(env,obj,sss_fid));

  c_message = (char *)((*env)->GetStringUTFChars(env,message,0));
  c_handle = (char *)((*env)->GetStringUTFChars(env,handle,0));

  rc = SSS_SendMessage(sss,c_handle,c_message);

  (*env)->ReleaseStringUTFChars(env,message,c_message);
  (*env)->ReleaseStringUTFChars(env,handle,c_handle);

  if (rc == -1){
    SSS_strerror(sss->sss_errno,errbuf,&errtype);
    switch (errtype){
    case SSS_SOCKET_ERROR:
    case SSS_COMMUNICATION_ERROR:
      ExceptionClass = (*env)->FindClass(env,"java/io/IOException");
      break;
    case SSS_INTERNAL_ERROR:
      ExceptionClass = (*env)->FindClass(env,"SSS_InternalError");
      break;
    }
    (*env)->ThrowNew(env,ExceptionClass,errbuf);
  }

  return rc;
}
  
JNIEXPORT jstring JNICALL Java_commlib_RecvMessage(JNIEnv *env,jobject obj,jstring handle){
  char *c_handle;
  jstring retval;
  char *buf;
  jfieldID sss_fid;
  jclass cls;
  SSS sss;
  jclass ExceptionClass;
  char errbuf[MAXERRLENGTH];
  int errtype;

  cls = (*env)->GetObjectClass(env,obj);

  sss_fid = (*env)->GetFieldID(env,cls,"sss_ptr","I");
			       
  sss = (SSS)((*env)->GetIntField(env,obj,sss_fid));

  c_handle = (char *)((*env)->GetStringUTFChars(env,handle,0));

  buf = SSS_RecvMessage(sss,c_handle);

  (*env)->ReleaseStringUTFChars(env,handle,c_handle);

  if (!buf){
    SSS_strerror(sss->sss_errno,errbuf,&errtype);
    switch (errtype){
    case SSS_SOCKET_ERROR:
    case SSS_COMMUNICATION_ERROR:
      ExceptionClass = (*env)->FindClass(env,"java/io/IOException");
      break;
    case SSS_INTERNAL_ERROR:
      ExceptionClass = (*env)->FindClass(env,"SSS_InternalError");
      break;
    }
    (*env)->ThrowNew(env,ExceptionClass,errbuf);
  }

  retval = (*env)->NewStringUTF(env,buf);

  free(buf);

  return retval;
}

JNIEXPORT void JNICALL Java_commlib_native_1finalize(JNIEnv *env,jobject obj){
  jfieldID sss_fid;
  jclass cls;
  SSS sss;

  printf("in finalize\n");

  cls = (*env)->GetObjectClass(env,obj);

  sss_fid = (*env)->GetFieldID(env,cls,"sss_ptr","I");

  sss = (SSS)((*env)->GetIntField(env,obj,sss_fid));

  SSS_Close(sss);

}

JNIEXPORT jstring JNICALL Java_commlib_get_1conf_1string(JNIEnv *env,jobject obj){
  jfieldID sss_fid;
  jclass cls;
  SSS sss;
  SSS_ConfigDict conf;
  char *retstr;
  jstring ret;

  cls = (*env)->GetObjectClass(env,obj);

  sss_fid = (*env)->GetFieldID(env,cls,"sss_ptr","I");

  sss = (SSS)((*env)->GetIntField(env,obj,sss_fid));

  conf = SSS_GetGlobalConfigDict(sss);

  retstr = malloc(snprintf(NULL,0,"%s:%s,%s:%s,%s:%s,%s:%s",
			   "sd_host",
			   SSS_GetFromConfigDict(conf,"sd_host"),
			   "sd_port",
			   SSS_GetFromConfigDict(conf,"sd_port"),
			   "sd_protocol",
			   SSS_GetFromConfigDict(conf,"sd_protocol"),
			   "debug",
			   SSS_GetFromConfigDict(conf,"debug")));

  sprintf(retstr,"%s:%s,%s:%s,%s:%s,%s:%s",
			   "sd_host",
			   SSS_GetFromConfigDict(conf,"sd_host"),
			   "sd_port",
			   SSS_GetFromConfigDict(conf,"sd_port"),
			   "sd_protocol",
			   SSS_GetFromConfigDict(conf,"sd_protocol"),
			   "debug",
			   SSS_GetFromConfigDict(conf,"debug"));
  
  ret = (*env)->NewStringUTF(env,retstr);

  SSS_FreeConfigDict(conf);

  free(retstr);

  return ret;
}

