#!/usr/bin/env python
# $Id: server.py 1.155 04/01/08 12:18:01-06:00 andrew@corvus.(none) $

from __future__ import nested_scopes

from cPickle import loads, dumps, load, dump
from atexit import register
from exceptions import Exception
from os import getpid, kill, wait, write, read
from select import select, error as selecterror
from signal import signal, SIGINT, SIGTERM, SIGCHLD
from socket import gethostname, SOL_SOCKET, SO_REUSEADDR, AF_INET, SOCK_STREAM, error as socketerr
import socket
from string import join, split
from sys import exit, exc_info
import sys
from syslog import openlog,syslog,closelog, LOG_INFO,LOG_ERR, LOG_LOCAL0
from syslogtb import enable as stbenable
from time import localtime, strftime, time
from xml.dom.minidom import parseString, Document
from xml.parsers.expat import ExpatError
from traceback import extract_tb
from types import ClassType
from restriction import restriction

from ssslib import comm_lib, CommunicationError, ConnectError, InitError, ProtocolError, SocketError, LookupError

class FakeValidationException:
    pass

try:
    from validate import validate,ValidationException
except ImportError:
    print "Warning: validation import failed.  XML Validation disabled."
    syslog(LOG_INFO,"Warning: validation import failed.  XML Validation disabled.")
    ValidationException = FakeValidationException

class StateLoadError(Exception):
    pass

class StateSaveError(Exception):
    pass

class SSSSelectServer:
    __statefields__=[]
    __validate__ = 1
    __timeout__ = 10
    
    def __init__(self,**kwargs):
        self.static=0
        self.kwargs=kwargs
        self.debug=kwargs.get('debug',0)
        self.comm=comm_lib(debug=self.debug)
        self.port=kwargs.get('port',0)
        if (("validate" in globals().keys()) and self.__validate__):
            self.validate=1
        else:
            self.validate=0
        prefs=self.comm.GetConfig()
        if not prefs['prefs']:
            self.protocol="challenge"
        else:
            self.protocol=kwargs.get('protocol',prefs['prefs'][0])
        self.__statefile__="/var/spool/sss/%s"%(self.__implementation__)
        self.schema_version=kwargs.get('schema_version','1.0')
        self.tier=kwargs.get('tier','1')
        self.__parseinput__()
        self.handles={}
        self.peers={}
        self.shut=0
        sys.excepthook = self.syslog_tb
        self.readsockets=[]
        self.socketdispatch={}
        self.__prebind__()
        for data in ['__implementation__','__component__','dispatch']:
            if not hasattr(self,data):
                raise TypeError, "%s not set after __prebind__"%(data)
        openlog(self.__implementation__,0,LOG_LOCAL0)
        try:
            d=self.comm.StaticLookup(self.__component__)
            self.comm.SetServerProtocol(d['protocol'])
            self.port=int(d['extra'])
            syslog(LOG_INFO, "Got static port %s for service %s"%(self.port,self.__component__))
            self.static=1
        except LookupError:
            pass
        # setup child procs
        self.children=self.__setupChildren__()
        # setup signal handlers
        signal(SIGINT,self.SigHand)
        signal(SIGTERM,self.SigHand)
        signal(SIGCHLD,self.SigChildHand)
        self.server=socket.socket(AF_INET,SOCK_STREAM)
        self.server.setsockopt(SOL_SOCKET,SO_REUSEADDR,1)
        self.server.bind(('',self.port))
        self.server.listen(32)
        self.port=self.server.getsockname()[1]
        syslog(LOG_INFO,"Bound to port %s"%(self.port))
        self.readsockets.append(self.server)
        self.socketdispatch[self.server]=self.AcceptNewConnection
        if self.__component__ != "service-directory":
            try:
                if not self.static:
                    self.comm.SDRegister(self.__component__,gethostname(),self.port,self.protocol,self.schema_version,self.tier)
            except CommunicationError, e:
                syslog(LOG_ERR, "Error Registering")
                self.shut=1
        self.__postbind__()
        if not hasattr(self,'schemabase'):
            self.schemabase=self.__component__
        # use atexit.register for cleanup
        register(self.Shutdown)

    def __parseinput__(self):
        pass

    def __prebind__(self):
        pass

    def __postbind__(self):
        pass

    def __setupChildren__(self):
        return []

    def __shutdown__(self):
        pass

    def syslog_tb(self,arg1,arg2,arg3):
        for line in extract_tb(arg3):
            syslog(LOG_ERR,'  File "%s", line %i, in %s'%(line[0],line[1],line[2]))
            print '  File "%s", line %i, in %s'%(line[0],line[1],line[2])
            syslog(LOG_ERR,'    %s'%line[3])
            print '    %s'%line[3]
        syslog(LOG_ERR,"%s: %s"%(arg1,arg2))
        print "%s: %s"%(arg1,arg2)

    def AcceptNewConnection(self,sock):
        (newsock,sockinfo)=sock.accept()
        try:
            h=self.comm.ServerInit(newsock)
            self.handles[newsock]=h
            self.readsockets.append(newsock)
            self.peers[newsock]=sockinfo
        except (CommunicationError,ProtocolError,SocketError), msg:
            syslog(LOG_ERR, "Failed to init %s: %s"%(sockinfo[0],msg))
        return 0

    def HandleSSSMessage(self,sock):
        messages=0
        (peer,peerport)=self.peers[sock]
        try:
            msg=self.comm.RecvMessage(self.handles[sock])
            messages+=1
        except (CommunicationError,SocketError,ProtocolError), msg:
            syslog(LOG_ERR,"%s from %s"%(msg,peer))
            syslog(LOG_ERR,"closing socket")
            handle = self.handles[sock]
            self.readsockets.remove(sock)
            del self.handles[sock]
            del self.peers[sock]
            self.comm.ServerClose(handle)
            return 0
            
        if msg == '':
            # we need to close up sock
            handle = self.handles[sock]
            self.readsockets.remove(sock)
            del self.handles[sock]
            del self.peers[sock]
            self.comm.ServerClose(handle)
            return 0
        # call handler here
        else:
            try:
                xmsg=parseString(msg)
            except:
                syslog(LOG_ERR,"parse error from %s: %s"%(peer,msg))
                self.comm.SendMessage(self.handles[sock],"<error type='Parse' msg='%s'/>"%msg)
                # we need to close up sock
                handle = self.handles[sock]
                self.readsockets.remove(sock)
                del self.handles[sock]
                del self.peers[sock]
                self.comm.ServerClose(handle)
                return 0

            cmd=xmsg.documentElement
            response = None
            try:
                func = self.dispatch[cmd.nodeName]
            except KeyError:
                syslog(LOG_ERR, "unhandled cmd %s"%(cmd.nodeName))
                response=Document().createElement("error")
                response.setAttribute("type",'Semantic')
                response.setAttribute("msg",'%s is an unknown request'%cmd.nodeName)
            if not response:
                try:
                    if self.validate:
                        validate(msg,"%s-in.xsd"%(self.schemabase))
                    # remove entraneous empty text nodes
                    for child in cmd.childNodes:
                        if child.nodeType == 3 and not child.nodeValue.strip():
                            cmd.removeChild(child)
                    response=func(cmd,(peer,peerport))
                except ValidationException:
                    (t,v,tb)=exc_info()
                    syslog(LOG_ERR,"Validation In:%s:"%msg)
                    syslog(LOG_ERR,str(v))
                    response=Document().createElement("error")
                    response.setAttribute("type",'Validation')
                    response.setAttribute("msg",str(v))
                except:
                    (t,v,tb)=exc_info()
                    response=Document().createElement("error")
                    response.setAttribute("type","Semantic")
                    response.setAttribute("msg","Unexpected command failure: %s: %s"%(t,v))
                    syslog(LOG_ERR, "Unexpected cmd failure")
                    syslog(LOG_ERR, "msg was: %s"%(msg))
                    for line in extract_tb(tb):
                        errstr = '  File "%s", line %i, in %s\n    %s\n'%line
                        syslog(LOG_ERR,errstr)
                    errstr = "%s: %s\n"%(t,v)
                    syslog(LOG_ERR,errstr)
                        
            # validate and send the response
            if self.validate:
                try:
                    validate(response.toxml(),"%s-out.xsd"%(self.schemabase))
                except ValidationException:
                    (t,v,tb)=exc_info()
                    syslog(LOG_ERR,"Validation Out:%s:"%response.toxml())
                    syslog(LOG_ERR,str(v))
                    
            try:
                self.comm.SendMessage(self.handles[sock],response.toxml())
            except (CommunicationError, ProtocolError, SocketError), msg:
                syslog(LOG_ERR, "%s from %s"%(msg,peer))
                handle = self.handles[sock]
                self.readsockets.remove(sock)
                del self.handles[sock]
                del self.peers[sock]
                self.comm.ServerClose(handle)
            return 1

    def PipeReadMsg(self,pipe):
        msglen=read(pipe,16)
        try:
            msgdata=read(pipe,int(msglen))
        except:
            syslog(LOG_ERR, "msglen format problem data: %s"%msglen)
        return loads(msgdata)

    def PipeWriteMsg(self,pipe,msg):
        msgdata=dumps(msg)
        write(pipe,"%16d"%(len(msgdata)))
        write(pipe,msgdata)

    def Serve(self):
        messages=0
        try:
            (inl,outl,other)=select(self.readsockets,[],[],self.__timeout__)
        except selecterror, s:
            if s[0] != 4:
                # ! EINTR
                # We got some other freaky error
                (t,v,tb)=exc_info()
                syslog(LOG_ERR, "Unexpected failure in select:")
                for line in extract_tb(tb):
                    errstr = '  File "%s", line %i, in %s\n    %s\n'%line
                    syslog(LOG_ERR,errstr)
                    errstr = "%s: %s\n"%(t,v)
                    syslog(LOG_ERR,errstr)
                    syslog(LOG_ERR,str(self.readsockets))
                    syslog(LOG_ERR,str(self.socketdispatch))
            return 0
        for sock in inl:
            if self.socketdispatch.has_key(sock):
                try:
                    messages+=self.socketdispatch[sock](sock)
                except:
                    (t,v,tb)=exc_info()
                    syslog(LOG_ERR, "Unexpected failure in socket dispatch:")
                    for line in extract_tb(tb):
                        errstr = '  File "%s", line %i, in %s\n    %s\n'%line
                        syslog(LOG_ERR,errstr)
                    errstr = "%s: %s\n"%(t,v)
                    syslog(LOG_ERR,errstr)
                    syslog(LOG_ERR, str(self.readsockets))
                    syslog(LOG_ERR, str(self.socketdispatch))
            else:
                messages+=self.HandleSSSMessage(sock)
        return messages
                
    def ServeN(self,num):
        done=0
        while done < num:
            done+=self.Serve()
            
    def ServeForever(self):
        while not self.shut:
            try:
                self.Serve()
            except selecterror:
                pass

    def SigHand(self,signum,frame):
        self.shut=1

    def SigChildHand(self,signum,frame):
        try:
            (pid,a)=wait()
            syslog(LOG_INFO, "pid %s exited"%(pid))
            if pid in self.children:
                self.children.remove(pid)
        except OSError:
            pass
            
    def Shutdown(self):
        self.__shutdown__()
        # Kill children
        map(lambda x:kill(x,SIGINT),self.children)
        self.server.close()
        syslog(LOG_INFO,"Shutting down")
        if self.__component__ != "service-directory":
            try:
                if not self.static:
                    self.comm.SDRemove(self.__component__,gethostname(),self.port,self.protocol)
            except:
                (t,v,tb)=exc_info()
                syslog(LOG_ERR, "Unexpected failure in sdremove:")
                for line in extract_tb(tb):
                    errstr = '  File "%s", line %i, in %s\n    %s\n'%line
                    syslog(LOG_ERR,errstr)
                errstr = "%s: %s\n"%(t,v)
                syslog(LOG_ERR,errstr)
        closelog()

    def ParseConfig(self):
        self.cf={}
        try:
            for data in open(self.cfname,'r').readlines():
                (var,value)=split(data,'=')
                self.cf[var]=value
        except IOError, e:
            syslog(LOG_ERR,"Unable to open conf file: ", e)

    def GetRefCount(self):
        d = {}
        # collect all classes
        for m in sys.modules.values():
            for sym in dir(m):
                o = getattr (m, sym)
                if type(o) is ClassType:
                    d[o] = sys.getrefcount (o)
        # sort by refcount
        pairs = map (lambda x: (x[1],x[0]), d.items())
        pairs.sort()
        pairs.reverse()
        return pairs
    
    def SaveState(self, depth=0):
        if depth > 3:
            raise StateSaveError
        state=tuple(map(lambda x:getattr(self,x),self.__statefields__))
        try:
            data=open(self.__statefile__,'w')
            dump(state,data)
            data.close()
        except (IOError,OSError):
            raise StateSaveError
        except RuntimeError:
            self.SaveState(depth+1)

    def LoadState(self):
        try:
            state=load(file(self.__statefile__))
        except (IOError,OSError,EOFError):
            raise StateLoadError
        if len(state) != len(self.__statefields__):
            raise StateLoadError
        map(lambda x:setattr(self,self.__statefields__[x],state[x]),xrange(len(state)))

class SubscriptionException(Exception):
    pass

class SSSSelectEventReceiver(SSSSelectServer):
    __implementation__="SSSSelectEventReceiver"
    __validate__=0

    def HandleXMLEvent(self, msg, (peer,port)):
        return Document().createElement('event-ok')

    def __parseinput__(self):
        self.__component__="er-%s-%s"%(getpid(),gethostname())
        self.scomponent=self.kwargs['component']
        self.msg=self.kwargs['msg']
        self.data=self.kwargs['data']
        self.dispatch={'events':self.HandleXMLEvent}

    def __postbind__(self):
        try:
            h=self.comm.ClientInit('event-manager')
            self.comm.SendMessage(h,"<add-subscription><subscription component='%s' msg='%s' data='%s' notify='%s'/></add-subscription>"%(self.scomponent,self.msg,self.data,self.__component__))
            self.comm.RecvMessage(h)
            self.comm.ClientClose(h)
        except:
            (t,v,tb)=exc_info()
            raise SubscriptionException, v

    def __shutdown__(self):
        # unsubscribe
        try:
            h=self.comm.ClientInit('event-manager')
            self.comm.SendMessage(h,"<del-subscription><subscription component='%s' msg='%s' data='%s' notify='%s'/></del-subscription>"%(self.scomponent,self.msg,self.data,self.__component__))
            self.comm.RecvMessage(h)
            self.comm.ClientClose(h)
        except:
            syslog(LOG_ERR, "Event Unsubscribe Failure.")

class SSSMultiEventReceiver(SSSSelectEventReceiver):
    __implementation__="MultiEventReceiver"
    __validate__=0
    
    def __parseinput__(self):
        if not hasattr(self,'__component__'):
            self.__component__="er-%s-%s"%(getpid(),gethostname())
        self.subs=[]
        self.dispatch={'events':self.HandleXMLEvent}

    def __postbind__(self):
        msg="<add-subscription><subscription component='%%s' msg='%%s' data='%%s' notify='%s'/></add-subscription>"%(self.__component__)
        try:
            h=self.comm.ClientInit('event-manager')
            for sub in self.subs:
                self.comm.SendMessage(h,msg%sub)
                self.comm.RecvMessage(h)
            self.comm.ClientClose(h)
        except:
            (t,v,vb)=exc_info()
            raise SubscriptionException, v

    def __shutdown__(self):
        msg="<del-subscription><subscription component='%%s' msg='%%s' data='%%s' notify='%s'/></del-subscription>"%(self.__component__)
        try:
            h=self.comm.ClientInit('event-manager')
            for sub in self.subs:
                self.comm.SendMessage(h,msg%sub)
                self.comm.RecvMessage(h)
            self.comm.ClientClose(h)
        except:
            syslog(LOG_ERR, "Event Unsubscribe Failure.")

def AddEvent(component,msg,data):
    comm = comm_lib()
    handle = comm.ClientInit('event-manager')
    response = Document()
    container=response.createElement("add-event")
    event = response.createElement("event")
    event.setAttribute("component",component)
    event.setAttribute("msg",msg)
    event.setAttribute("data",data)
    event.setAttribute('time',strftime("%c",localtime()))
    response.appendChild(container)
    response.documentElement.appendChild(event)
    comm.SendMessage(handle,response.toxml())
    response = comm.RecvMessage(handle)
    comm.ClientClose(handle)
    if len(parseString(response).documentElement.childNodes) != 1:
        raise "ResponseError"

def VPrint(results):
    lendata=map(lambda x:map(len,x),results)
    fsizes=tuple(map(lambda i:apply(max,map(lambda x:x[i],lendata)),range(len(lendata[0]))))
    fstring="%%%ss   %%%ss   %%%ss"%fsizes
    i=0
    for data in results:
        print fstring%data
        if not i:
            print "================================================================================"
            i=1

class rquery:
    def __init__(self,component,cursor,se,rc,table=None):
        self.component=component
        self.cursor=cursor
        self.subelement=se
        self.rcontainer=rc
        self.table=table
        self.handlers={}

    def Get(self,xe,(peer,port)):
        restrictions=map(restriction,xe.getElementsByTagName(self.subelement))
        for item in restrictions:
            if self.table:
                item.table=self.table
            self.cursor.execute(item.GetSQL())
            item.SetRows(self.cursor.fetchall())
        data=reduce(lambda x,y:x+y,map(lambda x:x.GetXML(),restrictions))
        response=Document().createElement(self.rcontainer)
        map(response.appendChild,data)
        return response

    def Add(self,xe,(peer,port)):
        restrictions=map(restriction,xe.getElementsByTagName(self.subelement))
        for item in restrictions:
            if self.table:
                item.table=self.table
            try:
                self.cursor.execute(item.AddSQL())
            except:
                response=Document().createElement('error')
                response.setAttribute('type','Data')
                return response
            self.cursor.execute(item.GetSQL())
            item.SetRows(self.cursor.fetchall())
        data=reduce(lambda x,y:x+y,map(lambda x:x.GetXML(),restrictions))
        response=Document().createElement(self.rcontainer)
        map(response.appendChild,data)
        return response

    def Del(self,xe,(peer,port)):
        restrictions=map(restriction,xe.getElementsByTagName(self.subelement))
        for item in restrictions:
            if self.table:
                item.table=self.table
            self.cursor.execute(item.GetSQL())
            item.SetRows(self.cursor.fetchall())
            self.cursor.execute(item.DelSQL())
        data=reduce(lambda x,y:x+y,map(lambda x:x.GetXML(),restrictions))
        response=Document().createElement(self.rcontainer)
        map(response.appendChild,data)
        return response

    def Set(self,xe,(peer,port)):
        restrictions=map(restriction,xe.getElementsByTagName(self.subelement))
        setsql=join(map(lambda x:"%s='%s'"%(x),xe._get_attributes().items()),',')
        for item in restrictions:
            if self.table:
                item.table = self.table
            if 'node' not in item.fields:
                item.fields.append('node')
            query=item.SetSQL()
            self.cursor.execute(query%(setsql))
            self.cursor.execute(item.GetSQL())
            item.SetRows(self.cursor.fetchall())
        events=Document().createElement('add-event')
        for item in restrictions:
            for node in item.GetField('node'):
                event=Document().createElement('event')
                event.setAttribute('component',self.component)
                event.setAttribute('msg', 'state-changed')
                event.setAttribute('data',node)
                event.setAttribute('time',strftime("%c",localtime()))
                events.appendChild(event)
        if events.childNodes:
            c=comm_lib()
            h=c.ClientInit('event-manager')
            c.SendMessage(h,events.toxml())
            r=c.RecvMessage(h)
            c.ClientClose(h)
        data=reduce(lambda x,y:x+y,map(lambda x:x.GetXML(),restrictions))
        response=Document().createElement(self.rcontainer)
        map(response.appendChild,data)
        return response

    def SetAction(self,name,fields,handler,signature):
        self.handlers[name]=(fields,handler,signature)
        setattr(self,name,self.Action)

    def Action(self,xe,(peer,port)):
        restrictions=map(restriction,xe.getElementsByTagName(self.subelement))
        (fields,handler,signature) = self.handlers[xe.tagName]
        args=map(lambda x:xe.getAttribute(x),signature)
        for item in restrictions:
            for field in fields:
                if field not in item.fields: item.fields.append(field)
            self.cursor.execute(item.GetSQL())
            item.SetRows(self.cursor.fetchall())
            apply(handler,(item.GetFields(fields),)+tuple(args))
        data=reduce(lambda x,y:x+y,map(lambda x:x.GetXML(),restrictions))
        response=Document().createElement(self.rcontainer)
        map(response.appendChild,data)
        return response

class rclient:
    def __init__(self,component,datatype,attrlist,argv):
        self.component=component
        self.datatype=datatype
        self.attrlist=attrlist
        getargre = recompile("^(-d ){0,1}get -r (?P<restrictions>(\w+=(\w|\.|\*)+ {0,1})*)$")
        setargre = recompile("^(-d ){0,1}set (?P<update>\w+=\w+) -r (?P<restrictions>(\w+=(\w|\.|\*)+ {0,1})*)$")
        addargre = recompile("^(-d ){0,1}add (?P<values>(\w+=(\w|\.|\*)+ {0,1})+)$")
        delargre = recompile("^(-d ){0,1}del -r (?P<restrictions>(\w+=(\w|\.|\*)+ {0,1})*)$")
        argstring = join(argv[1:]," ")

        creator=Document()
        if getargre.match(argstring):
            regex=getargre
        elif setargre.match(argstring):
            regex=setargre
        elif addargre.match(argstring):
            regex=addargre
        elif delargre.match(argstring):
            regex=delargre
        else:
            self.Usage()
        
        wrapper=creator.createElement("get-%s"%(datatype))
        restrictions=regex.match(argstring).group("restrictions")
        data=creator.createElement(datatype)
        map(lambda x:data.setAttribute(x,'*'), attrlist)
        for rest in split(restrictions):
            apply(data.setAttribute,tuple(split(rest,'=')))

        if setargre.match(argstring):
            apply(wrapper.setAttribute,tuple(split(setargre.match(argstring).group('update'),'=')))
            
        wrapper.append(data)
        creator.appendChild(wrapper)
    
        sss = comm_lib(debug=('-d' in argv))
        handle = sss.ClientInit(component)
        sss.SendMessage(handle,creator.toxml())
        msg = sss.RecvMessage(handle)
        sss.ClientClose(handle)
        del sss
        try:
            domobj = parseString(msg)
        except:
            print "server error"
            exit(1)

        self.dom=domobj

    def Display(self):
        if self.dom.getElementsByTagName(self.datatype):
            fields=dom.getElementsByTagName(self.datatype)[0]._attrs.keys()
            results=[tuple(fields)]
            for record in dom.getElementsByTagName(self.datatype):
                results.append(tuple(map(lambda x:record.getAttribute(x), fields)))
            lendata=map(lambda x:map(len,x),results)
            fsizes=tuple(map(lambda i:apply(max,map(lambda x:x[i],lendata)),range(len(lendata[0]))))
            fstring=join(['%%%ss']*len(fsizes),'   ')%(fsizes)
            i=0
            for data in results:
                print fstring%data
                if not i:
                    print "================================================================================"
                    i=1
    

class TimerException(Exception):
    pass 

class Timer:
    '''A timer class'''
    def __init__(self,name):
        self.num_intervals = 0
        self.total_time    = 0
        self.name          = name
        self.min           = 1000000
        self.max           = 0
        self.starttime     = -1

    def start(self):
        self.starttime = time()

    def stop(self):
        self.stoptime = time()
        self.new_interval = self.stoptime - self.starttime
        self.total_time += self.new_interval
        self.num_intervals += 1
        self.avg_interval = self.total_time / self.num_intervals
        if self.new_interval < self.min:
            self.min = self.new_interval
        if self.new_interval > self.max:
            self.max = self.new_interval
        self.starttime = -1               # reset

    def check(self):
        if self.starttime == -1:
            raise TimerException
        else:
            return time() - self.starttime

    def tprint(self):
        print 'statistics for %s timer:' % (self.name)
        print 'total time          = %6.3f seconds' % (self.total_time)
        print 'number of intervals =  %s' % (self.num_intervals)
        print 'average interval    = %6.3f seconds' % (self.avg_interval)
        print 'max interval        = %6.3f seconds' % (self.max)
        print 'min interval        = %6.3f seconds' % (self.min)

